"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports["default"] = throttle;

var _requestAnimationFrameTimeout = require("request-animation-frame-timeout");

// For some weird reason, in Chrome, `setTimeout()` would lag up to a second (or more) behind.
// Turns out, Chrome developers have deprecated `setTimeout()` API entirely without asking anyone.
// Replacing `setTimeout()` with `requestAnimationFrame()` can work around that Chrome bug.
// https://github.com/bvaughn/react-virtualized/issues/722

/**
 * Same as `lodash`'s `throttle()` for functions with no arguments.
 * @param  {function} func
 * @param  {number} interval
 * @return {function}
 */
function throttle(func, interval) {
  var timeout;
  var executedAt = 0;

  var scheduled = function scheduled() {
    timeout = undefined;
    executedAt = Date.now();
    func();
  };

  return function () {
    var now = Date.now();
    var remaining = interval - (now - executedAt);

    if (remaining <= 0) {
      if (timeout) {
        (0, _requestAnimationFrameTimeout.clearTimeout)(timeout);
        timeout = undefined;
      }

      executedAt = now;
      func();
    } else if (!timeout) {
      timeout = (0, _requestAnimationFrameTimeout.setTimeout)(scheduled, remaining);
    }
  };
}
//# sourceMappingURL=throttle.js.map